# Copyright (c) 2019 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Handle migrations of signac schema versions."""

import os
import sys
from contextlib import contextmanager

from filelock import FileLock
from packaging import version

from ...common.config import get_config
from ...version import SCHEMA_VERSION, __version__
from .v0_to_v1 import migrate_v0_to_v1

FN_MIGRATION_LOCKFILE = ".SIGNAC_PROJECT_MIGRATION_LOCK"


MIGRATIONS = {
    ("0", "1"): migrate_v0_to_v1,
}


def _reload_project_config(project):
    project_reloaded = project.get_project(
        root=project.root_directory(), search=False, _ignore_schema_version=True
    )
    project._config = project_reloaded._config


def _update_project_config(project, **kwargs):
    """Update the project configuration."""
    for fn in ("signac.rc", ".signacrc"):
        config = get_config(project.fn(fn))
        if "project" in config:
            break
    else:
        raise RuntimeError("Unable to determine project configuration file.")
    config.update(kwargs)
    config.write()
    _reload_project_config(project)


@contextmanager
def _lock_for_migration(project):
    lock = FileLock(project.fn(FN_MIGRATION_LOCKFILE))
    try:
        with lock:
            yield
    finally:
        try:
            os.unlink(lock.lock_file)
        except FileNotFoundError:
            pass


def _collect_migrations(project):
    schema_version = version.parse(SCHEMA_VERSION)

    def get_config_schema_version():
        return version.parse(project._config["schema_version"])

    if get_config_schema_version() > schema_version:
        # Project config schema version is newer and therefore not supported.
        raise RuntimeError(
            "The signac schema version used by this project is {}, but signac {} "
            "only supports up to schema version {}. Try updating signac.".format(
                get_config_schema_version(), __version__, SCHEMA_VERSION
            )
        )

    while get_config_schema_version() < schema_version:
        for (origin, destination), migration in MIGRATIONS.items():
            if version.parse(origin) == get_config_schema_version():
                yield (origin, destination), migration
                break
        else:
            raise RuntimeError(
                "The signac schema version used by this project is {}, but signac {} "
                "uses schema version {} and does not know how to migrate.".format(
                    get_config_schema_version(), __version__, schema_version
                )
            )


def apply_migrations(project):
    """Apply migrations to a project."""
    with _lock_for_migration(project):
        for (origin, destination), migrate in _collect_migrations(project):
            try:
                print(
                    f"Applying migration for version {origin} to {destination}... ",
                    end="",
                    file=sys.stderr,
                )
                migrate(project)
            except Exception as e:
                raise RuntimeError(f"Failed to apply migration {destination}.") from e
            else:
                _update_project_config(project, schema_version=destination)
                print("OK", file=sys.stderr)
                yield origin, destination


__all__ = [
    "apply_migrations",
]



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Wrapper around json parsing library."""
import logging
from json import JSONEncoder, load, loads
from json.decoder import JSONDecodeError
from typing import Any, Dict, Optional

from deprecation import deprecated

from ..version import __version__

logger = logging.getLogger(__name__)

try:
    import numpy

    NUMPY = True
except ImportError:
    NUMPY = False

"""
THIS MODULE IS DEPRECATED!
"""


# this class is deprecated
class CustomJSONEncoder(JSONEncoder):
    """Attempt to JSON-encode objects beyond the default supported types.

    This encoder will attempt to obtain a JSON-serializable representation of
    an object that is otherwise not serializable, by calling the object's
    `_as_dict()` method.
    """

    @deprecated(
        deprecated_in="1.7",
        removed_in="2.0",
        current_version=__version__,
        details="The core.json module bundled with signac is deprecated.",
    )
    def default(self, o: Any) -> Dict[str, Any]:
        if NUMPY:
            if isinstance(o, numpy.number):
                return o.item()
            elif isinstance(o, numpy.ndarray):
                return o.tolist()
        try:
            return o._as_dict()
        except AttributeError:
            # Call the super method, which raises a TypeError if it cannot
            # encode the object.
            return super().default(o)


@deprecated(
    deprecated_in="1.7",
    removed_in="2.0",
    current_version=__version__,
    details="The core.json module bundled with signac is deprecated.",
)
def dumps(o: Any, sort_keys: bool = False, indent: Optional[int] = None) -> str:
    """Convert a JSON-compatible mapping into a string."""
    return CustomJSONEncoder(sort_keys=sort_keys, indent=indent).encode(o)


__all__ = ["loads", "load", "dumps", "JSONDecodeError"]
